from PIL import Image


def load_img(_image: Image.Image, path: str) -> Image.Image:
    """
    Load image from disk.

    :param _image: input image (ignored).
    :param path: path to image to load.
    :returns: loaded image.
    """
    return Image.open(path).convert("RGBA")



from typing import Union

from PIL.Image import Image


def fit(  # noqa: WPS210
    image: Image,
    width: Union[str, int],
    height: Union[str, int],
) -> Image:
    """
    Fit image into given dimensions.

    This processor resizes image,
    and keep aspect ratio of an image.

    :param image: image to fit.
    :param width: desired width.
    :param height: desired height.
    :return: resized image.
    """
    width = int(width)
    height = int(height)

    width_factor = width / image.width
    height_factor = height / image.height

    scale_factor = max(width_factor, height_factor)

    resized = image.resize(
        (int(image.width * scale_factor), int(image.height * scale_factor)),
    )

    return resized.crop(
        (
            (resized.width - width) // 2,
            (resized.height - height) // 2,
            (resized.width + width) // 2,
            (resized.height + height) // 2,
        ),
    )



from typing import Union

from PIL import Image


def blank_image(
    _image: Image.Image,
    width: Union[str, int],
    height: Union[str, int],
    color: str = "#000000",
) -> Image.Image:
    """
    Generate blank image with one color.

    :param _image: imnput image (ignored)
    :param width: width of a new image.
    :param height: height of a new image.
    :param color: color of an image, defaults to "#000000"
    :returns: created image.
    """
    width = int(width)
    height = int(height)
    return Image.new("RGBA", (width, height), color)



from typing import Optional

from PIL.Image import Image


def resize(
    image: Image,
    width: Optional[str] = None,
    height: Optional[str] = None,
    factor: Optional[str] = None,
) -> Image:
    """
    Resize image with given size.

    :param image: target image to resize.
    :param width: new image's width
    :param height: new image's height
    :param factor: scale factor.
        Image dimensions will be multiplied by this parameter.
        if passed, width and heigh parameters are ignored.
    :return: resized image.
    """
    new_width = int(width or image.width)
    new_height = int(height or image.height)

    if factor is not None:
        new_width = int(image.width * float(factor))
        new_height = int(image.height * float(factor))

    return image.resize((new_width, new_height))



from functools import partial
from multiprocessing import Pool
from typing import Optional, Tuple, Union

from loguru import logger
from PIL import Image

from music_bg.config import Layer
from music_bg.context import Context


def process_layer(  # noqa: WPS210
    image: Image.Image,
    context: Context,
    layer: Layer,
) -> Tuple[Union[str, int], Image.Image]:
    """
    Process image layer.

    This function sequentially applies image processors
    from the configuration on an album cover.

    :param image: Album cover.
    :param context: Current MBG context.
    :param layer: Current layer.

    :raises ValueError: if unknown variable was used in config.

    :return: Name of the layer and processed image.
    """
    image = image.copy()
    for processor in layer.processors:
        processor_func = context.get_processor(processor.name)
        logger.debug(f"Applying {processor.name} on layer {layer.name}")
        if not processor.args:
            image = processor_func(image)
            continue

        arguments = {}
        for arg, arg_value in processor.args.items():
            try:
                arguments[arg] = str(arg_value).format_map(context.variables())
            except KeyError as kerr:
                raise ValueError(f'Unknown variable "{{{kerr.args[0]}}}"') from kerr

        image = processor_func(image, **arguments)

    return layer.name, image


def process_image(  # noqa: WPS210
    image: Image.Image,
    context: Context,
) -> Optional[Image.Image]:  # : WPS210
    """
    Process album cover according to the config.

    This function processes every layer from configuration
    and merges it.

    :param image: album cover.
    :param context: current music_bg context.
    :raises ValueError: if layer image bigger than the screen.
    :returns: processed image.
    """
    if not context.config.layers:
        return None

    blender = context.config.blender
    if not blender:
        blender = [layer.name for layer in context.config.layers]

    layers = Pool().map(partial(process_layer, image, context), context.config.layers)
    layers_map = dict(layers)

    image = Image.new("RGBA", (context.screen.width, context.screen.height))
    for blend_index in blender:
        overlay_img = layers_map[blend_index]
        if overlay_img.height > image.height or overlay_img.width > image.width:
            raise ValueError("Layer image bigger than biggest screen.")
        image.alpha_composite(
            overlay_img,
            (
                (image.width - overlay_img.width) // 2,
                (image.height - overlay_img.height) // 2,
            ),
        )
    return image



import subprocess

from loguru import logger

from music_bg.context import Context


def set_background(filename: str, context: Context) -> None:
    """
    Update current background.

    :param filename: path to background file.
    :param context: current mbg context.
    """
    logger.debug("Setting background")
    command = context.config.set_command.format(filename)
    try:
        subprocess.run(["/bin/sh", "-c", command]).check_returncode()  # noqa: S603
    except subprocess.CalledProcessError as exc:
        logger.exception(exc)


def reset_background(context: Context) -> None:
    """
    Return background to the default state.

    :param context: current mbg context.
    """
    logger.debug("Reseting background")
    command = context.config.reset_command
    try:
        subprocess.run(["/bin/sh", "-c", command]).check_returncode()  # noqa: S603
    except subprocess.CalledProcessError as exc:
        logger.exception(exc)



import dbus
from dbus.mainloop.glib import DBusGMainLoop
from gi.repository import GLib
from loguru import logger

from music_bg.context import Context
from music_bg.dbus.handlers import player_exit_handler, player_signal_handler


def run_loop(context: Context) -> None:
    """
    Run dbus listener loop.

    :param context: current context.
    """
    logger.info("Setting up dbus connection.")
    dbus_loop = DBusGMainLoop()
    bus = dbus.SessionBus(mainloop=dbus_loop)
    bus.add_signal_receiver(
        player_signal_handler(context),
        dbus_interface="org.freedesktop.DBus.Properties",
        path="/org/mpris/MediaPlayer2",
        interface_keyword="dbus_interface",
        arg0="org.mpris.MediaPlayer2.Player",
    )
    bus.add_signal_receiver(
        player_exit_handler(context),
        dbus_interface="org.freedesktop.DBus",
        signal_name="NameOwnerChanged",
        interface_keyword="dbus_interface",
    )
    logger.info("Loop is ready.")
    loop = GLib.MainLoop()
    loop.run()



from PIL.Image import Image


def noop(image: Image) -> Image:
    """
    Dummy processor.

    It does nothing and returns the same image.

    :param image: input image
    :return: same image
    """
    return image



from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser, Namespace
from pathlib import Path

from xdg import xdg_config_home


def parse_args() -> Namespace:
    """
    Parse CLI arguments.

    :return: parsed Namespace.
    """
    parser = ArgumentParser(
        formatter_class=ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "-c",
        "--config",
        help="Path to the configuration file",
        default=xdg_config_home() / ".mbg.json",
        type=Path,
        dest="config_path",
    )

    parser.add_argument(
        "-V",
        "--version",
        help="Show application version",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "-r",
        "--reload",
        help="Reload configuration before setting background",
        action="store_true",
        default=False,
    )

    subparsers = parser.add_subparsers(dest="subparser_name")

    gen_parser = subparsers.add_parser(
        "gen",
        help="Generate config",
    )

    info_parser = subparsers.add_parser(
        "info",
        help="Get additional configuration information",
    )

    info_parser.add_argument(
        "-p",
        "--processors",
        action="store_true",
        help="Show available processors",
        dest="show_processors",
    )

    info_parser.add_argument(
        "-v",
        "--vars",
        action="store_true",
        help="Show available variables",
        dest="show_vars",
    )

    gen_parser.add_argument(
        "-c",
        "--config",
        help="Path to the configuration file",
        default=xdg_config_home() / ".mbg.json",
        type=Path,
        dest="config_path",
    )

    return parser.parse_args()



from sys import stdout

from loguru import logger

from music_bg.config import LogLevel


def init_logger(level: LogLevel) -> None:
    """Configure music_bg logging.

    :param level: New log level.
    """
    logger.remove()
    logger.add(
        sink=stdout,
        level=level.value,
        diagnose=True,
        backtrace=True,
    )



import subprocess
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional

import entrypoints
from loguru import logger
from PIL.Image import Image
from pydantic import BaseModel, Field

from music_bg.config import Config


class Metadata(BaseModel):
    """Music metadata."""

    track_id: Optional[str] = Field(alias="mpris:trackid")
    art_url: Optional[str] = Field(alias="mpris:artUrl")
    album: Optional[str] = Field(alias="xesam:album")
    artists: Optional[List[str]] = Field(alias="xesam:artist")
    title: Optional[str] = Field(alias="xesam:title")
    track_number: Optional[int] = Field(alias="xesam:trackNumber")
    url: Optional[str] = Field(alias="xesam:url")


class Screen(BaseModel):
    """Screen dimensions model."""

    width: int = 1366
    height: int = 768


class Context:  # noqa: WPS230
    """music_bg context object."""

    def __init__(
        self,
        config_path: Optional[Path] = None,
        reloadable: Optional[bool] = None,
    ):
        """
        Method used to initialize class.

        This is a good topic to discuss, but I've decided to
        create this method, because __init__ method is called
        everytime you create a class instance. But this method
        is called only once in __new__ method.

        :param config_path: path to the config.
        :param reloadable: if config must be reloadable.
        """
        self.config_path = config_path or Path("~/.mbg.json")
        self.config = Config()
        self.reloadable = reloadable or False
        self.last_status = ""
        self.screen = Screen()
        self.metdata = Metadata()
        self.src_image = None
        self.processors_map: Dict[str, Callable[..., Image]] = {}
        self.variables_providers: Dict[str, Callable[..., Any]] = {
            "screen": Context.get_screen_size,
            "metadata": Context.get_metadata,
        }
        self.reload()

    def reload(self) -> None:
        """Perform full context reload."""
        self.reload_config()
        self.reload_processors()
        self.reload_screen_size()
        self.reload_variables_providers()

    def reload_config(self) -> None:
        """Update configuration from file."""
        self.config = Config.from_file(self.config_path.expanduser())

    def reload_screen_size(self) -> None:
        """
        Update biggest screen size.

        :raises ValueError: if can't get screen size
            or format is invalid.
        """
        logger.debug("Updating screen resolution")
        try:
            run_result = subprocess.run(  # noqa: S603
                ["/bin/sh", "-c", self.config.screen_resolution_command],
                stdout=subprocess.PIPE,
            )
            run_result.check_returncode()
        except subprocess.CalledProcessError:
            raise ValueError("Can't read screen resolution.")

        resolution = run_result.stdout.decode().strip()
        logger.debug(f"Found resolution: {resolution}")
        try:
            width, height = map(int, resolution.split("x"))
        except ValueError:
            raise ValueError(f"Invalid resolution format: '{resolution}'.")

        self.screen = Screen(
            width=width,
            height=height,
        )

    def reload_processors(self) -> None:
        """Find and load in memory all image processors."""
        for entrypoint in entrypoints.get_group_all("mbg_processors"):
            processor_func = entrypoint.load()
            self.processors_map[entrypoint.name] = processor_func

    def reload_variables_providers(self) -> None:
        """Find and load in memory all variables providers."""
        for entrypoint in entrypoints.get_group_all("mbg_variables"):
            variable_func = entrypoint.load()
            self.variables_providers[entrypoint.name] = variable_func

    def get_processor(self, processor_name: str) -> Callable[..., Image]:
        """
        Get processor function by name.

        :param processor_name: name of processor.
        :raises ValueError: if processor is not found.
        :return: processor function.
        """
        if processor_name not in self.processors_map:
            logger.error(f"Processor {processor_name} is not found.")
            raise ValueError(f"Unknown processor {processor_name}")
        return self.processors_map[processor_name]

    def get_screen_size(self) -> Screen:
        """
        Get dimensions of the biggest screen.

        screen has two fields:
        * width
        * height

        You can use it as

        {screen.width} or {screen.height}

        :returns: a Screen object.
        """
        return self.screen

    def get_metadata(self) -> Metadata:
        """
        Get current track metadata.

        :return: metadata info.
        """
        return self.metdata

    def variables(self) -> Dict[str, Any]:
        """
        Get variables mapping.

        :return: mapping with variables.
        """
        return {
            name: var_proc(self) for name, var_proc in self.variables_providers.items()
        }



from tempfile import NamedTemporaryFile
from typing import Any, Callable, Dict, Optional

import requests
from loguru import logger
from PIL import Image

from music_bg.background import reset_background, set_background
from music_bg.context import Context, Metadata
from music_bg.img_processors.processor import process_image


def guard_metadata(context: Context, player_args: Dict[str, Any]) -> Optional[Metadata]:
    """
    This function constructs Metadata.

    If this request was already processed,
    None would be returned.

    :param context: current mbg context.
    :param player_args: arguents passed to dbus.
    :returns: metadata.
    """
    status = player_args.get("PlaybackStatus")
    raw_meta = player_args.get("Metadata")
    if raw_meta is None:
        return None
    metadata = Metadata(**raw_meta)
    if (  # noqa: WPS337
        str(metadata.track_id) == str(context.metdata.track_id)
        and str(status) == context.last_status
    ):
        return None
    if metadata.art_url is None:
        logger.debug("Can't get art_url")
        return None
    context.last_status = str(status)
    context.metdata = metadata
    return metadata


def player_signal_handler(  # noqa: WPS213, WPS210, C901
    context: Context,
) -> Callable[..., None]:
    """
    Dbus handler generator.

    :param context: current context.
    :return: dbus listener function.
    """

    def _player_signal_handler(  # noqa: WPS213, WPS210, C901
        _dbus_interface: str,
        player_args: Dict[str, Any],
        *_args: Any,
        **_kwargs: Dict[str, Any],
    ) -> None:
        """
        This signal is triggered when player's properties are changed.

        It checks for song's metadata and sets it as the wallpaper.

        :param _dbus_interface: name of the interface on which event apeared.
        :param player_args: current state of a player.
        :param _args: dbus additional arguments.
        :param _kwargs: additional dbus info.
        """
        status = player_args.get("PlaybackStatus")
        metadata = guard_metadata(context, player_args)
        if metadata is None:
            return
        if status != "Playing":
            reset_background(context)
            return
        logger.debug(f"Requesting {metadata.art_url}")
        response = requests.get(str(metadata.art_url), stream=True)
        if not response.ok:
            logger.debug(f"Image response returned status {response.status_code}")
            return
        context.reload_screen_size()

        if context.reloadable:
            context.reload()

        image = Image.open(response.raw).convert("RGBA")
        context.src_image = image.copy()
        processed = process_image(image, context)
        if processed is not None:
            with NamedTemporaryFile(mode="wb", delete=True) as temp_file:
                processed.save(temp_file, format="png")
                logger.debug(f"Background saved at {temp_file.name}")
                set_background(temp_file.name, context)

    return _player_signal_handler


def player_exit_handler(context: Context) -> Callable[..., None]:
    """
    Dbus handler generator.

    :param context: current context.
    :return: dbus listener function.
    """

    def _player_exit_handler(
        name: str,
        _old_name: str,
        new_name: str,
        **_kwargs: Dict[str, Any],
    ) -> None:
        """
        Function which called when someone leaves dbus.

        :param name: name of the interface.
        :param _old_name: old dbus name.
        :param new_name: new dbus name.
        :param _kwargs: different kwargs.
        """
        if str(name).startswith("org.mpris.MediaPlayer2") and str(new_name) == "":
            logger.info(f"Player {name} exited")
            reset_background(context)

    return _player_exit_handler



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['music_bg', 'music_bg.dbus', 'music_bg.img_processors']

package_data = \
{'': ['*']}

install_requires = \
['Pillow-SIMD>=7.0.0,<8.0.0',
 'PyGObject>=3.40.1,<4.0.0',
 'dbus-python>=1.2.16,<2.0.0',
 'entrypoints>=0.3,<0.4',
 'loguru>=0.5.3,<0.6.0',
 'pydantic>=1.8.2,<2.0.0',
 'requests>=2.26.0,<3.0.0',
 'toml>=0.10.2,<0.11.0',
 'xdg>=5.1.1,<6.0.0']

entry_points = \
{'console_scripts': ['music_bg = music_bg.__main__:main'],
 'mbg_processors': ['blank_image = '
                    'music_bg.img_processors.blank_img:blank_image',
                    'fit = music_bg.img_processors.fit:fit',
                    'load_img = music_bg.img_processors.load_img:load_img',
                    'noop = music_bg.img_processors.noop:noop',
                    'resize = music_bg.img_processors.resize:resize']}

setup_kwargs = {
    'name': 'music-bg',
    'version': '0.2.0',
    'description': 'Dynamic wallpapers with power of mpris2.',
    'long_description': '![python version](https://img.shields.io/pypi/pyversions/music_bg?style=flat-square) ![Build status](https://img.shields.io/github/workflow/status/music-bg/music_bg/Release%20package?style=flat-square) [![version](https://img.shields.io/pypi/v/music_bg?style=flat-square)](https://pypi.org/project/music_bg/)\n\n\n<div align="center">\n<img src="https://raw.githubusercontent.com/music-bg/music_bg/master/images/logo.png" width=700>\n</div>\n\n<p align="center">\n    <a href="https://github.com/music-bg/music_bg/wiki">Wiki</a> •\n    <a href="#about">About</a> •\n    <a href="#installation">Installation</a>\n</p>\n\n<div align="center">\n<h1 id="about">About</h1>\n</div>\n\nThis project is a dynamic wallpaper changer.\nIt waits untill you turn on the music,\ndownloads album cover if it\'s possible and\nsets it as your wallpapper.\n\nAnd the main thing, this project is highly customizable.\nYou can install plugins and write your own configurations.\n\n<div align="center">\n<h1 id="installation">Installation</h1>\n</div>\n\n⚠️ This project requires DBUS to be installed. Currently only linux is supported. ⚠️\n\nTo install this project simply run:\n```bash\npip install music_bg\n# Also you can install extra processors\n# by running\npip install music_bg_extra\n```\n\nThat\'s it. Now you can read our [wiki](https://github.com/music-bg/music_bg/wiki)\nto set up your own music background.\n',
    'author': 'Pavel Kirilin',
    'author_email': 'win10@list.ru',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/music-bg/music_bg/wiki',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import inspect
from importlib import metadata
from pathlib import Path

from loguru import logger
from PIL.Image import Image

from music_bg.argparse import parse_args
from music_bg.config import Config
from music_bg.context import Context
from music_bg.dbus.loop import run_loop
from music_bg.logging import init_logger


def generate_config(config_path: Path) -> None:
    """
    Generate default config file.

    :param config_path: path to config.
    """
    if config_path.exists():
        print(f"Config {config_path} already exists")
        return
    Config().save(config_path)
    print(f"Config successfully generated at {config_path}")


def show_version() -> None:
    """Show installed version of a Music Background."""
    version = metadata.version("music_bg")
    print(f"Music background v{version}")


def print_processors(context: Context) -> None:  # noqa: WPS213, WPS210, WPS210, C901
    """
    Print information about available image processors.

    :param context: music_bg context.
    """
    print(" Processors ".center(80, "#"))

    for name, func in context.processors_map.items():
        print("-" * 80)
        print(f"name: {name}")
        print("type: processor")
        args = inspect.signature(func)
        custom_args = []
        for parameter in args.parameters.values():
            if parameter.annotation != Image:
                custom_args.append(parameter)
        if custom_args:
            print()
            print(" args ".center(20, "="))
            for arg in custom_args:
                arg_info = f"* {arg.name}"
                if arg.annotation != arg.empty:
                    arg_info = f"{arg_info}: {arg.annotation}"
                if arg.default == arg.empty:
                    arg_info = f"{arg_info} (required)"
                print(arg_info)
        doc = inspect.getdoc(func)
        if doc is not None:
            print()
            print(" doc ".center(20, "="))
            print(doc)


def print_variables(context: Context) -> None:
    """
    Print information about available variables.

    :param context: current mbg context.
    """
    print(" Variables ".center(80, "#"))

    for name, func in context.variables_providers.items():
        print("-" * 80)
        print(f"name: {name}")
        print("type: variable")
        doc = inspect.getdoc(func)
        if doc is not None:
            print()
            print(" doc ".center(20, "="))
            print(doc)


def show_info(
    context: Context,
    show_processors: bool = False,
    show_variables: bool = False,
) -> None:
    """
    Show information about current context.

    This function shows available processors
    and variable providers.

    :param context: mbg context.
    :param show_processors: show information about processors.
    :param show_variables: show information about variables.
    """
    show_version()
    if show_processors:
        print()
        print_processors(context)
    if show_variables:
        print()
        print_variables(context)


def main() -> None:
    """The main entrypoint of a program."""
    logger.remove()
    args = parse_args()
    if args.version:
        show_version()
        return
    if args.subparser_name == "gen":
        generate_config(args.config_path)
        return
    context = Context(args.config_path, args.reload)
    if args.subparser_name == "info":
        show_info(
            context,
            show_processors=args.show_processors,
            show_variables=args.show_vars,
        )
        return
    init_logger(context.config.log_level)
    logger.debug(f"Using config {args.config_path}")
    try:
        run_loop(context)
    except KeyboardInterrupt:
        logger.info("Goodbye!")


if __name__ == "__main__":
    main()



import enum
import json
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Optional, Union

import toml
from pydantic import BaseModel


class ImageProcessor(BaseModel):
    """Image processor config."""

    name: str
    args: Optional[Dict[str, Any]]


class Layer(BaseModel):
    """Mergeable layer."""

    name: Union[str, int]
    processors: List[ImageProcessor]


class LogLevel(enum.Enum):
    """Possible log levels."""

    INFO = "INFO"  # noqa: WPS110
    ERROR = "ERROR"
    DEBUG = "DEBUG"


class Config(BaseModel):
    """User configuration object."""

    blender: List[Union[str, int]] = []
    log_level: LogLevel = LogLevel.INFO

    set_command: str = 'feh --bg-fill "{0}"'
    reset_command: str = "nitrogen --restore"

    screen_resolution_command: str = "xrandr | grep '*' | cut -d' ' -f4 | sort --human-numeric-sort --reverse | head -n 1"  # noqa: E501
    layers: List[Layer] = []

    @classmethod
    def get_serde_by_extension(  # noqa: WPS234
        cls,
        ext: str,
    ) -> Iterable[Callable[..., Dict[str, Any]]]:
        """
        Get serializer and deserializer for the given file extension.

        :param ext: file extension.
        :raises ValueError: if unknown extension was passed.
        :return: serializer and deserializer.
        """
        extension_map = {
            "json": (json.dump, json.load),
            "toml": (toml.dump, toml.load),
        }
        if ext not in extension_map:
            known_formats = ",".join(extension_map.keys())
            raise ValueError(
                f"Unknown config format. Supported formats: {known_formats}",
            )
        return extension_map[ext]  # type: ignore

    @classmethod
    def from_file(
        cls,
        config_path: Path,
    ) -> "Config":
        """
        Load config from a given path.

        :param config_path: path to the configuration file.
        :return: Config instance.
        """
        extension = config_path.name.split(".")[-1]

        _, des = cls.get_serde_by_extension(extension)

        with open(config_path, "r") as config_io:
            deserialized = des(config_io)
        return cls(**deserialized)

    def save(self, config_path: Path) -> None:
        """Write config to file.

        :param config_path: path to file.
        """
        extension = config_path.name.split(".")[-1]

        ser, _ = self.get_serde_by_extension(extension)

        with open(config_path, "w") as config_io:
            ser(self.dict(exclude={"log_level"}), config_io)
